// Written by Yi Yang, North Carolina State University, Sep. 2010.
// All rights reserved.

// includes, system
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

// includes, project
#include <cutil.h>
#include <cublas.h>

// includes, kernels
#include <mm_kernel.cu>


void
computeGold(float* C, const float* A, const float* B, int aWIDTH, int aHEIGHT, int bWIDTH, int bHEIGHT, int cWIDTH, int cHEIGHT)
{
    for (int j = 0; j < cWIDTH; ++j) {
    	for (int k=0; k<cHEIGHT; k++)
    	{
			float sum = 0;
			for (int i = 0; i < aWIDTH; ++i) {
				float a = A[k * aWIDTH + i];
				float b = B[i*bWIDTH+j];
				sum += a * b;
			}
			C[j+cWIDTH*k] = sum;
    	}
    }
}

void printUsage() {
}



void randomInit(float* data, int size)
{
    for (int i = 0; i < size; ++i)
        data[i] = rand() / (float)RAND_MAX;
}

template<int width>
void mmkernel(int aWIDTH, int aHEIGHT, int bWIDTH, cudaStream_t *stream, dim3 grid, dim3 threads, float* kbuff, float* d_inputB, float* d_outputC, int kMemSize) {
	int numberX =  aWIDTH/CONST_WIDTH;
	int cWIDTH = bWIDTH;
	for (int j=0; j<aWIDTH; j+=CONST_WIDTH) {
		for (int i=0; i<aHEIGHT; i+=NUM_STREAM*CONST_HEIGHT) {
			if (j==0) {
				CUDA_SAFE_CALL(cudaMemcpyToSymbolAsync(kernel_matrix_0, kbuff+CONST_WIDTH*CONST_HEIGHT*(numberX*(i/CONST_HEIGHT+0)+j/CONST_WIDTH), kMemSize, 0, cudaMemcpyDeviceToDevice,  stream[0]));
				mm_const_kernel<0, 0, width><<< grid, threads, 0, stream[0]>>>(d_inputB+j*bWIDTH, d_outputC+cWIDTH*(CONST_HEIGHT*0+i));

				CUDA_SAFE_CALL(cudaMemcpyToSymbolAsync(kernel_matrix_1, kbuff+CONST_WIDTH*CONST_HEIGHT*(numberX*(i/CONST_HEIGHT+1)+j/CONST_WIDTH), kMemSize, 0, cudaMemcpyDeviceToDevice,  stream[1]));
				mm_const_kernel<1, 0, width><<< grid, threads, 0, stream[1]>>>(d_inputB+j*bWIDTH, d_outputC+cWIDTH*(CONST_HEIGHT*1+i));

				CUDA_SAFE_CALL(cudaMemcpyToSymbolAsync(kernel_matrix_2, kbuff+CONST_WIDTH*CONST_HEIGHT*(numberX*(i/CONST_HEIGHT+2)+j/CONST_WIDTH), kMemSize, 0, cudaMemcpyDeviceToDevice,  stream[2]));
				mm_const_kernel<2, 0, width><<< grid, threads, 0, stream[2]>>>(d_inputB+j*bWIDTH, d_outputC+cWIDTH*(CONST_HEIGHT*2+i));

				CUDA_SAFE_CALL(cudaMemcpyToSymbolAsync(kernel_matrix_3, kbuff+CONST_WIDTH*CONST_HEIGHT*(numberX*(i/CONST_HEIGHT+3)+j/CONST_WIDTH), kMemSize, 0, cudaMemcpyDeviceToDevice,  stream[3]));
				mm_const_kernel<3, 0, width><<< grid, threads, 0, stream[3]>>>(d_inputB+j*bWIDTH, d_outputC+cWIDTH*(CONST_HEIGHT*3+i));
			}
			else {
				CUDA_SAFE_CALL(cudaMemcpyToSymbolAsync(kernel_matrix_0, kbuff+CONST_WIDTH*CONST_HEIGHT*(numberX*(i/CONST_HEIGHT+0)+j/CONST_WIDTH), kMemSize, 0, cudaMemcpyDeviceToDevice,  stream[0]));
				mm_const_kernel<0, 1, width><<< grid, threads, 0, stream[0]>>>(d_inputB+j*bWIDTH, d_outputC+cWIDTH*(CONST_HEIGHT*0+i));

				CUDA_SAFE_CALL(cudaMemcpyToSymbolAsync(kernel_matrix_1, kbuff+CONST_WIDTH*CONST_HEIGHT*(numberX*(i/CONST_HEIGHT+1)+j/CONST_WIDTH), kMemSize, 0, cudaMemcpyDeviceToDevice,  stream[1]));
				mm_const_kernel<1, 1, width><<< grid, threads, 0, stream[1]>>>(d_inputB+j*bWIDTH, d_outputC+cWIDTH*(CONST_HEIGHT*1+i));

				CUDA_SAFE_CALL(cudaMemcpyToSymbolAsync(kernel_matrix_2, kbuff+CONST_WIDTH*CONST_HEIGHT*(numberX*(i/CONST_HEIGHT+2)+j/CONST_WIDTH), kMemSize, 0, cudaMemcpyDeviceToDevice,  stream[2]));
				mm_const_kernel<2, 1, width><<< grid, threads, 0, stream[2]>>>(d_inputB+j*bWIDTH, d_outputC+cWIDTH*(CONST_HEIGHT*2+i));

				CUDA_SAFE_CALL(cudaMemcpyToSymbolAsync(kernel_matrix_3, kbuff+CONST_WIDTH*CONST_HEIGHT*(numberX*(i/CONST_HEIGHT+3)+j/CONST_WIDTH), kMemSize, 0, cudaMemcpyDeviceToDevice,  stream[3]));
				mm_const_kernel<3, 1, width><<< grid, threads, 0, stream[3]>>>(d_inputB+j*bWIDTH, d_outputC+cWIDTH*(CONST_HEIGHT*3+i));
			}

		}
	}
}


void test(int x, int y, int z, int blocksize) {

	int aWIDTH = x;
	int aHEIGHT = y;
	int bWIDTH = z;
	int bHEIGHT = x;
	int cWIDTH = z;
	int cHEIGHT = y;



    cublasStatus status;
    status = cublasInit();
    if (status != CUBLAS_STATUS_SUCCESS) {
        fprintf (stderr, "!!!! CUBLAS initialization error\n");
		exit (1);
    }


    unsigned int aMemSize = sizeof(float) * aWIDTH*aHEIGHT;
    unsigned int bMemSize = sizeof(float) * bWIDTH*bHEIGHT;
    unsigned int cMemSize = sizeof(float) * cWIDTH*cHEIGHT;

	// allocate host memory
    float* inputA = (float*)malloc(aMemSize);
    float* inputB = (float*)malloc(bMemSize);
    float* outputC = (float*)malloc(cMemSize);
	float* ref = (float*) malloc( cMemSize);


	printf("Initialize input, A (%d, %d), B (%d, %d), C (%d, %d)\n", aWIDTH, aHEIGHT, bWIDTH, bHEIGHT, cWIDTH, cHEIGHT);

	// allocate device memory
	float* d_outputC;
    cudaMalloc( (void**) &d_outputC, cMemSize);
	float* d_inputA;
    cudaMalloc( (void**) &d_inputA, aMemSize);
    float* d_inputB;
    cudaMalloc( (void**) &d_inputB, bMemSize);

	float* d_inputK;
    cudaMalloc( (void**) &d_inputK, aMemSize);



    // assign input value
    randomInit(inputA, aWIDTH*aHEIGHT);
    randomInit(inputB, bWIDTH*bHEIGHT);

	unsigned int timer = 0;




    cudaStream_t *stream;
	stream = (cudaStream_t *)malloc((NUM_STREAM) * sizeof(cudaStream_t));
    for(int i=0; i<NUM_STREAM; i++)
    {
        cudaStreamCreate(&stream[i]);
    }


	float kerneltime = 0;


	CUDA_SAFE_CALL( cudaMemcpy(d_inputA, inputA, aMemSize,
							cudaMemcpyHostToDevice) );
	CUDA_SAFE_CALL( cudaMemcpy(d_inputB, inputB, bMemSize,
							cudaMemcpyHostToDevice) );

	int counter = 20;

    {

		cudaMemset(d_outputC, 0, cMemSize);


		for (int k=0; k<counter; k++) {
			CUT_SAFE_CALL(cutCreateTimer(&timer));
			cudaThreadSynchronize();
			CUT_SAFE_CALL(cutStartTimer(timer));
			// execute the kernel
			cublasSgemm('n', 'n', bWIDTH, aHEIGHT, aWIDTH, 1.0f, d_inputB,
					bWIDTH, d_inputA, aWIDTH, 0.0f, d_outputC, bWIDTH);
			// stop and destroy timer
			cudaThreadSynchronize();
			kerneltime += cutGetTimerValue(timer);
			CUT_SAFE_CALL(cutStopTimer(timer));
			CUT_SAFE_CALL(cutDeleteTimer(timer));
		}

		// check if kernel execution generated and error
		CUT_CHECK_ERROR("Kernel execution failed");

		// copy result from device to host
		CUDA_SAFE_CALL(cudaMemcpy(ref, d_outputC, cMemSize,
								  cudaMemcpyDeviceToHost) );
		kerneltime = kerneltime/counter;

		printf("cublasSgemm Processing time: %f (ms), %f Gflops \n", kerneltime, 2.0*aWIDTH*aHEIGHT*bWIDTH/kerneltime/1e6);
		printf("ignore cpu, using cublas as reference\n");
    }

	{
		cudaMemset(d_outputC, 0, cMemSize);
		kerneltime = 0;
    	float transpose_time = 0;
        unsigned int kMemSize = sizeof(float) * CONST_WIDTH*CONST_HEIGHT;



    	kerneltime = 0;
		cutCreateTimer(&timer);
		cudaThreadSynchronize();
		CUT_SAFE_CALL(cutStartTimer(timer));

		float* kbuff = d_inputK;
		{
			int bsize = 16;
			int tx = bsize;
			int ty = bsize;
			int bx = aWIDTH/bsize;
			int by = aHEIGHT/bsize;

			dim3 threads(tx, ty);
			dim3 grid(bx, by);
			input_transpose<<< grid, threads>>>(d_inputA, d_inputK, aWIDTH);
		}
		cudaThreadSynchronize();
		cutStopTimer(timer);
		transpose_time = cutGetTimerValue(timer);
	    CUT_SAFE_CALL(cutDeleteTimer(timer));




		int tx = blocksize;
		int ty = 1;
		int bx = cWIDTH/tx;
		int by = 1;

		dim3 threads(tx, ty);
		dim3 grid(bx, by);
		for (int k=0; k<counter; k++) {
			cutCreateTimer(&timer);
			cudaThreadSynchronize();
			CUT_SAFE_CALL(cutStartTimer(timer));

			switch (bWIDTH)
			{
			case 4096:
				mmkernel<4096>(aWIDTH, aHEIGHT, bWIDTH, stream, grid, threads, kbuff, d_inputB, d_outputC, kMemSize);
				break;
			case 5120:
				mmkernel<5120>(aWIDTH, aHEIGHT, bWIDTH, stream, grid, threads, kbuff, d_inputB, d_outputC, kMemSize);
				break;
			case 6144:
				mmkernel<6144>(aWIDTH, aHEIGHT, bWIDTH, stream, grid, threads, kbuff, d_inputB, d_outputC, kMemSize);
				break;
			case 7168:
				mmkernel<7168>(aWIDTH, aHEIGHT, bWIDTH, stream, grid, threads, kbuff, d_inputB, d_outputC, kMemSize);
				break;
			case 8192:
				mmkernel<8192>(aWIDTH, aHEIGHT, bWIDTH, stream, grid, threads, kbuff, d_inputB, d_outputC, kMemSize);
				break;
			
			}

			cudaStreamSynchronize(stream[0]);
			cudaStreamSynchronize(stream[1]);
			cudaStreamSynchronize(stream[2]);
			cudaStreamSynchronize(stream[3]);


			cutStopTimer(timer);
			kerneltime += cutGetTimerValue(timer);
			CUT_SAFE_CALL(cutDeleteTimer(timer));
		}


		kerneltime = kerneltime/counter;
		CUDA_SAFE_CALL(cudaMemcpy(outputC, d_outputC, cMemSize,
								  cudaMemcpyDeviceToHost) );
//		printf("transpose Processing time: %f (ms), %f GB/S \n", transpose_time, 2.0*4*aHEIGHT*aWIDTH/transpose_time/1e6);
		printf("matmul_const Processing time: %f, %f (ms), %f Gflops, %f Gflops \n", transpose_time, kerneltime, 2.0*aWIDTH*aHEIGHT*bWIDTH/kerneltime/1e6, 2.0*aWIDTH*aHEIGHT*bWIDTH/(transpose_time+kerneltime)/1e6);
		CUTBoolean res = cutCompareL2fe(outputC, ref, cWIDTH*cHEIGHT, 1e-6f);
		printf("Test %s \n", (1 == res) ? "PASSED" : "FAILED");
	}





	free(inputA);
	free(inputB);
	free(outputC);
    free(ref);
    cudaFree(d_inputA);
    cudaFree(d_inputK);
    cudaFree(d_inputB);
    cudaFree(d_outputC);

    for(int i=0; i<NUM_STREAM; i++)
    {
        cudaStreamDestroy(stream[i]);
    }
    free(stream);


    status = cublasShutdown();
    if (status != CUBLAS_STATUS_SUCCESS) {
        fprintf (stderr, "!!!! shutdown error\n");
    }
}


////////////////////////////////////////////////////////////////////////////////
// Program main
////////////////////////////////////////////////////////////////////////////////
int main( int argc, char** argv)
{
	int blocksize = 128;
	if (argc==2)
	{
		blocksize = atoi(argv[1]);
	}	
    CUT_DEVICE_INIT(1, "");


	for (int i = 4096; i<=1024*8; i+=1024)
		for (int j=4096; j<=8*1024; j+=1024)
			test(i, i, j, blocksize);
    CUT_EXIT(1, "");
}
