// Written by Yi Yang, North Carolina State University, Sep. 2010.
// All rights reserved.

#define CONST_WIDTH 	128
#define CONST_HEIGHT 	16


#define NUM_STREAM 4


__constant__ float kernel_matrix_0[CONST_WIDTH*CONST_HEIGHT];
__constant__ float kernel_matrix_1[CONST_WIDTH*CONST_HEIGHT];
__constant__ float kernel_matrix_2[CONST_WIDTH*CONST_HEIGHT];
__constant__ float kernel_matrix_3[CONST_WIDTH*CONST_HEIGHT];



#define STORE() {output[0] = sum[offset++]; output += W_IN_KERNEL;}

#define LOAD() { sum[offset++] = output[0]; output += W_IN_KERNEL;}

#define LOAD_ZERO() { sum[offset++] = 0; }

#define COMPUTE_ONE_2(kernel_matrix, id) {\
	sum[id] += b*kernel_matrix[i*CONST_HEIGHT+id];\
}


#define COMPUTE_16(kernel_matrix) {\
COMPUTE_ONE_2(kernel_matrix, 0);\
COMPUTE_ONE_2(kernel_matrix, 1);\
COMPUTE_ONE_2(kernel_matrix, 2);\
COMPUTE_ONE_2(kernel_matrix, 3);\
COMPUTE_ONE_2(kernel_matrix, 4);\
COMPUTE_ONE_2(kernel_matrix, 5);\
COMPUTE_ONE_2(kernel_matrix, 6);\
COMPUTE_ONE_2(kernel_matrix, 7);\
COMPUTE_ONE_2(kernel_matrix, 8);\
COMPUTE_ONE_2(kernel_matrix, 9);\
COMPUTE_ONE_2(kernel_matrix, 10);\
COMPUTE_ONE_2(kernel_matrix, 11);\
COMPUTE_ONE_2(kernel_matrix, 12);\
COMPUTE_ONE_2(kernel_matrix, 13);\
COMPUTE_ONE_2(kernel_matrix, 14);\
COMPUTE_ONE_2(kernel_matrix, 15);\
}


template<int constid, int it, int width>
__global__	void mm_const_kernel(float* input, float* output)
{
#undef W_IN_KERNEL
#define W_IN_KERNEL (width)


	int x = blockIdx.x*blockDim.x+threadIdx.x;
	{

		float sum[CONST_HEIGHT];

		input += x;
		output += x;
		if (it!=0)
		{
			int offset = 0;
			LOAD(); LOAD(); LOAD(); LOAD();
			LOAD(); LOAD(); LOAD(); LOAD();
			LOAD(); LOAD(); LOAD(); LOAD();
			LOAD(); LOAD(); LOAD(); LOAD();
			output -= W_IN_KERNEL*16;
		}
		else {
			int offset = 0;
			LOAD_ZERO(); LOAD_ZERO(); LOAD_ZERO(); LOAD_ZERO();
			LOAD_ZERO(); LOAD_ZERO(); LOAD_ZERO(); LOAD_ZERO();
			LOAD_ZERO(); LOAD_ZERO(); LOAD_ZERO(); LOAD_ZERO();
			LOAD_ZERO(); LOAD_ZERO(); LOAD_ZERO(); LOAD_ZERO();
		}

		#pragma unroll
		for (int i=0; i<CONST_WIDTH; i++) {
			float b = input[0];
			if (constid==0) {COMPUTE_16(kernel_matrix_0);}
			if (constid==1) {COMPUTE_16(kernel_matrix_1);}
			if (constid==2) {COMPUTE_16(kernel_matrix_2);}
			if (constid==3) {COMPUTE_16(kernel_matrix_3);}
			input += W_IN_KERNEL;
		}

		{
			int offset = 0;
			STORE(); STORE(); STORE(); STORE();
			STORE(); STORE(); STORE(); STORE();
			STORE(); STORE(); STORE(); STORE();
			STORE(); STORE(); STORE(); STORE();
		}
	}
}


//transpose 16x16 sub block into one array
__global__	void input_transpose(float* input, float* output, int width)
{
	__shared__ float shared[16][17];
	int bx = blockIdx.x;
	int by = blockIdx.y;
	int x = threadIdx.x;
	int y = threadIdx.y;


	output +=(by*(width/16)+bx)*16*16;
	input += (by*16+y)*width+bx*16;
	shared[y][x] = input[x];
	__syncthreads();
	output[y*16+x] = shared[x][y];
}


